@@ -1,25 +1,32 @@
 // BadWolf: Minimalist and privacy-oriented WebKitGTK+ browser
-// Copyright © 2019-2020 Badwolf Authors <https://hacktivis.me/projects/badwolf>
+// SPDX-FileCopyrightText: 2019-2023 Badwolf Authors <https://hacktivis.me/projects/badwolf>
 // SPDX-License-Identifier: BSD-3-Clause
 
 #include "badwolf.h"
 
+#include "bookmarks.h"
 #include "config.h"
 #include "downloads.h"
+#include "fmt.h"
 #include "keybindings.h"
 #include "uri.h"
+#include "userscripts.h"
 
+#include <assert.h>
 #include <glib/gi18n.h>   /* _() and other internationalization/localization helpers */
 #include <libsoup/soup.h> /* soup* */
 #include <locale.h>       /* LC_* */
-#include <stdio.h>        /* perror(), fprintf() */
+#include <stdio.h>        /* perror(), fprintf(), snprintf() */
 #include <stdlib.h>       /* malloc() */
 #include <unistd.h>       /* access() */
 
-gchar *web_extensions_directory;
 const gchar *homepage = "https://hacktivis.me/projects/badwolf";
 const gchar *version  = VERSION;
 
+static gchar *web_extensions_directory;
+static uint64_t context_id_counter = 0;
+GtkTreeModel *bookmarks_completion_model;
+
 static gboolean WebViewCb_close(WebKitWebView *webView, gpointer user_data);
 static gboolean WebViewCb_web_process_terminated(WebKitWebView *webView,
                                                  WebKitWebProcessTerminationReason reason,
@@ -66,11 +73,11 @@ static void new_tabCb_clicked(GtkButton *new_tab, gpointer user_data);
 static void closeCb_clicked(GtkButton *close, gpointer user_data);
 static void
 notebookCb_switch__page(GtkNotebook *notebook, GtkWidget *page, guint page_num, gpointer user_data);
+void content_managerCb_ready(GObject *store, GAsyncResult *result, gpointer user_data);
 
 static gboolean
-WebViewCb_close(WebKitWebView *webView, gpointer user_data)
+WebViewCb_close(WebKitWebView *UNUSED(webView), gpointer user_data)
 {
-	(void)webView;
 	struct Client *browser = (struct Client *)user_data;
 
 	gtk_widget_destroy(browser->box);
@@ -81,11 +88,10 @@ WebViewCb_close(WebKitWebView *webView, gpointer user_data)
 }
 
 static gboolean
-WebViewCb_web_process_terminated(WebKitWebView *webView,
+WebViewCb_web_process_terminated(WebKitWebView *UNUSED(webView),
                                  WebKitWebProcessTerminationReason reason,
                                  gpointer user_data)
 {
-	(void)webView;
 	struct Client *browser = (struct Client *)user_data;
 
 	switch(reason)
@@ -107,14 +113,19 @@ WebViewCb_web_process_terminated(WebKitWebView *webView,
 }
 
 static gboolean
-WebViewCb_notify__uri(WebKitWebView *webView, GParamSpec *pspec, gpointer user_data)
+WebViewCb_notify__uri(WebKitWebView *UNUSED(webView), GParamSpec *UNUSED(pspec), gpointer user_data)
 {
-	(void)webView;
-	(void)pspec;
 	const gchar *location_uri;
 	struct Client *browser = (struct Client *)user_data;
 
 	location_uri = webkit_web_view_get_uri(browser->webView);
+	printf("location_uri: <%s>\n", location_uri);
+
+	// Don't set if location_uri is NULL / empty, latter happens on WebProcess termination
+	if(location_uri == NULL || location_uri[0] == '\0')
+	{
+		return TRUE;
+	}
 
 	gtk_entry_set_text(GTK_ENTRY(browser->location), location_uri);
 
@@ -129,17 +140,30 @@ WebViewCb_notify__uri(WebKitWebView *webView, GParamSpec *pspec, gpointer user_d
 GtkWidget *
 badwolf_new_tab_box(const gchar *title, struct Client *browser)
 {
-	(void)browser;
+	/* flawfinder: ignore. bound checks are done */
+	char context_id_str[BADWOLF_CTX_SIZ] = {0, 0, 0, 0, 0, 0, 0};
+
+	fmt_context_id(browser->context_id, context_id_str);
+
 	GtkWidget *tab_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
 	gtk_widget_set_name(tab_box, "browser__tabbox");
+
 	GtkWidget *close =
 	    gtk_button_new_from_icon_name("window-close-symbolic", GTK_ICON_SIZE_LARGE_TOOLBAR);
 	gtk_widget_set_name(close, "browser__tabbox__close");
 	GtkWidget *label = gtk_label_new(title);
 	gtk_widget_set_name(label, "browser__tabbox__label");
+	GtkWidget *label_event_box = gtk_event_box_new();
+	gtk_container_add(GTK_CONTAINER(label_event_box), label);
+	GtkWidget *context_label = gtk_label_new(context_id_str);
+	gtk_widget_set_name(context_label, "browser__tabbox__context_label");
+	GtkWidget *context_label_event_box = gtk_event_box_new();
+	gtk_container_add(GTK_CONTAINER(context_label_event_box), context_label);
 	GtkWidget *playing =
 	    gtk_image_new_from_icon_name("audio-volume-high-symbolic", GTK_ICON_SIZE_SMALL_TOOLBAR);
 	gtk_widget_set_name(playing, "browser__tabbox__playing");
+	GtkWidget *playing_event_box = gtk_event_box_new();
+	gtk_container_add(GTK_CONTAINER(playing_event_box), playing);
 
 #ifdef BADWOLF_TAB_BOX_WIDTH
 	gtk_widget_set_size_request(label, BADWOLF_TAB_BOX_WIDTH, -1);
@@ -151,9 +175,11 @@ badwolf_new_tab_box(const gchar *title, struct Client *browser)
 
 	gtk_label_set_ellipsize(GTK_LABEL(label), BADWOLF_TAB_LABEL_ELLIPSIZE);
 	gtk_label_set_single_line_mode(GTK_LABEL(label), TRUE);
+	gtk_label_set_single_line_mode(GTK_LABEL(context_label), TRUE);
 
-	gtk_box_pack_start(GTK_BOX(tab_box), playing, FALSE, FALSE, 0);
-	gtk_box_pack_start(GTK_BOX(tab_box), label, TRUE, TRUE, 0);
+	gtk_box_pack_start(GTK_BOX(tab_box), context_label_event_box, TRUE, TRUE, 0);
+	gtk_box_pack_start(GTK_BOX(tab_box), playing_event_box, FALSE, FALSE, 0);
+	gtk_box_pack_start(GTK_BOX(tab_box), label_event_box, TRUE, TRUE, 0);
 	gtk_box_pack_start(GTK_BOX(tab_box), close, FALSE, FALSE, 0);
 
 	gtk_button_set_relief(GTK_BUTTON(close), GTK_RELIEF_NONE);
@@ -164,27 +190,38 @@ badwolf_new_tab_box(const gchar *title, struct Client *browser)
 
 	gtk_widget_show_all(tab_box);
 	gtk_widget_set_visible(playing, webkit_web_view_is_playing_audio(browser->webView));
+	gtk_widget_set_events(label, GDK_BUTTON_RELEASE_MASK);
+	g_signal_connect(
+	    tab_box, "button-release-event", G_CALLBACK(tab_boxCb_button_release_event), browser);
 
 	return tab_box;
 }
 
 static gboolean
-WebViewCb_notify__title(WebKitWebView *webView, GParamSpec *pspec, gpointer user_data)
+WebViewCb_notify__title(WebKitWebView *UNUSED(webView),
+                        GParamSpec *UNUSED(pspec),
+                        gpointer user_data)
 {
-	(void)webView;
-	(void)pspec;
 	struct Client *browser = (struct Client *)user_data;
 
+	// Don't set if title is NULL / empty, latter happens on WebProcess crash
+	const char *title = webkit_web_view_get_title(browser->webView);
+	printf("title: <%s>\n", title);
+	if(title == NULL || title[0] == '\0')
+	{
+		return TRUE;
+	}
+
 	webView_tab_label_change(browser, NULL);
 
 	return TRUE;
 }
 
 static gboolean
-WebViewCb_notify__is__playing__audio(WebKitWebView *webView, GParamSpec *pspec, gpointer user_data)
+WebViewCb_notify__is__playing__audio(WebKitWebView *UNUSED(webView),
+                                     GParamSpec *UNUSED(pspec),
+                                     gpointer user_data)
 {
-	(void)webView;
-	(void)pspec;
 	struct Client *browser = (struct Client *)user_data;
 
 	webView_tab_label_change(browser, NULL);
@@ -197,11 +234,11 @@ webView_tab_label_change(struct Client *browser, const gchar *title)
 {
 	GtkWidget *notebook = browser->window->notebook;
 
-#define title_IS_EMPTY (title == NULL) || strnlen(title, 2) == 0
+#define title_IS_EMPTY title == NULL || title[0] == '\0'
 
 	if(title_IS_EMPTY) title = webkit_web_view_get_title(browser->webView);
 	if(title_IS_EMPTY) title = webkit_web_view_get_uri(browser->webView);
-	if(title_IS_EMPTY) title = "BadWolf";
+	if(title_IS_EMPTY) title = _("Empty Title");
 
 	gtk_notebook_set_tab_label(
 	    GTK_NOTEBOOK(notebook), browser->box, badwolf_new_tab_box(title, browser));
@@ -214,12 +251,10 @@ webView_tab_label_change(struct Client *browser, const gchar *title)
 }
 
 static gboolean
-WebViewCb_notify__estimated_load_progress(WebKitWebView *webView,
-                                          GParamSpec *pspec,
+WebViewCb_notify__estimated_load_progress(WebKitWebView *UNUSED(webView),
+                                          GParamSpec *UNUSED(pspec),
                                           gpointer user_data)
 {
-	(void)webView;
-	(void)pspec;
 	struct Client *browser = (struct Client *)user_data;
 	gdouble progress;
 
@@ -233,13 +268,11 @@ WebViewCb_notify__estimated_load_progress(WebKitWebView *webView,
 }
 
 static gboolean
-WebViewCb_mouse_target_changed(WebKitWebView *webView,
+WebViewCb_mouse_target_changed(WebKitWebView *UNUSED(webView),
                                WebKitHitTestResult *hit,
-                               guint modifiers,
+                               guint UNUSED(modifiers),
                                gpointer user_data)
 {
-	(void)webView;
-	(void)modifiers;
 	struct Client *browser = (struct Client *)user_data;
 
 	if(webkit_hit_test_result_context_is_link(hit))
@@ -255,9 +288,8 @@ WebViewCb_mouse_target_changed(WebKitWebView *webView,
 }
 
 static gboolean
-WebViewCb_scroll_event(GtkWidget *widget, GdkEvent *event, gpointer data)
+WebViewCb_scroll_event(GtkWidget *UNUSED(widget), GdkEvent *event, gpointer data)
 {
-	(void)widget;
 	struct Client *browser = (struct Client *)data;
 	gdouble delta_x, delta_y;
 	gdouble zoom;
@@ -276,44 +308,67 @@ WebViewCb_scroll_event(GtkWidget *widget, GdkEvent *event, gpointer data)
 
 static WebKitWebView *
 WebViewCb_create(WebKitWebView *related_web_view,
-                 WebKitNavigationAction *navigation_action,
+                 WebKitNavigationAction *UNUSED(navigation_action),
                  gpointer user_data)
 {
-	(void)navigation_action;
-	struct Window *window  = (struct Window *)user_data;
-	struct Client *browser = new_browser(window, NULL, related_web_view);
+	struct Client *old_browser = (struct Client *)user_data;
+	struct Client *browser     = NULL;
 
-	if(badwolf_new_tab(GTK_NOTEBOOK(window->notebook), browser, FALSE) < 0)
+	// shouldn't be needed but better be safe
+	old_browser->webView = related_web_view;
+
+	browser = new_browser(old_browser->window, NULL, old_browser);
+
+	if(badwolf_new_tab(GTK_NOTEBOOK(old_browser->window->notebook), browser, FALSE) < 0)
 		return NULL;
 	else
 		return browser->webView;
 }
 
 static gboolean
-WebViewCb_permission_request(WebKitWebView *web_view,
+WebViewCb_permission_request(WebKitWebView *UNUSED(web_view),
                              WebKitPermissionRequest *request,
-                             gpointer user_data)
+                             gpointer UNUSED(user_data))
 {
-	(void)web_view;
-	(void)user_data;
-
 	webkit_permission_request_deny(request);
 
 	return TRUE; /* Stop other handlers */
 }
 
 static gboolean
-WebViewCb_decide_policy(WebKitWebView *web_view,
+WebViewCb_decide_policy(WebKitWebView *UNUSED(web_view),
                         WebKitPolicyDecision *decision,
                         WebKitPolicyDecisionType decision_type,
                         gpointer user_data)
 {
+	struct Client *old_browser = (struct Client *)user_data;
 	WebKitResponsePolicyDecision *r;
-	(void)web_view;
-	(void)user_data;
+	WebKitNavigationPolicyDecision *n;
+	WebKitNavigationAction *navigation_action;
 
 	switch(decision_type)
 	{
+	case WEBKIT_POLICY_DECISION_TYPE_NAVIGATION_ACTION:
+		n                 = WEBKIT_NAVIGATION_POLICY_DECISION(decision);
+		navigation_action = webkit_navigation_policy_decision_get_navigation_action(n);
+
+		if(GDK_CONTROL_MASK == webkit_navigation_action_get_modifiers(navigation_action) ||
+		   2 == webkit_navigation_action_get_mouse_button(navigation_action))
+		{
+			WebKitURIRequest *uri   = webkit_navigation_action_get_request(navigation_action);
+			const gchar *target_url = webkit_uri_request_get_uri(uri);
+			struct Client *browser  = new_browser(old_browser->window, target_url, old_browser);
+
+			badwolf_new_tab(GTK_NOTEBOOK(browser->window->notebook), browser, FALSE);
+			webkit_web_view_load_uri(browser->webView, target_url);
+			webkit_policy_decision_ignore(decision);
+		}
+		else
+		{
+			/* Use whatever default there is. */
+			return FALSE;
+		}
+		break;
 	case WEBKIT_POLICY_DECISION_TYPE_RESPONSE:
 		r = WEBKIT_RESPONSE_POLICY_DECISION(decision);
 		if(!webkit_response_policy_decision_is_mime_type_supported(r))
@@ -330,10 +385,10 @@ WebViewCb_decide_policy(WebKitWebView *web_view,
 }
 
 static void
-WebViewCb_load_changed(WebKitWebView *webView, WebKitLoadEvent load_event, gpointer user_data)
+WebViewCb_load_changed(WebKitWebView *UNUSED(webView),
+                       WebKitLoadEvent UNUSED(load_event),
+                       gpointer user_data)
 {
-	(void)webView;
-	(void)load_event;
 	struct Client *browser = (struct Client *)user_data;
 
 	gtk_widget_set_sensitive(browser->back, webkit_web_view_can_go_back(browser->webView));
@@ -375,19 +430,21 @@ detail_tls_certificate_flags(GTlsCertificateFlags tls_errors)
 }
 
 static gboolean
-WebViewCb_load_failed_with_tls_errors(WebKitWebView *web_view,
+WebViewCb_load_failed_with_tls_errors(WebKitWebView *UNUSED(web_view),
                                       gchar *failing_text,
                                       GTlsCertificate *certificate,
                                       GTlsCertificateFlags errors,
                                       gpointer user_data)
 {
-	(void)web_view;
-	(void)certificate;
-	(void)errors;
 	struct Client *browser = (struct Client *)user_data;
 	gchar *error_details   = detail_tls_certificate_flags(errors);
 	gint dialog_response;
+
+#ifndef USE_LIBSOUP2
+	GUri *failing_uri = g_uri_parse(failing_text, G_URI_FLAGS_NONE, NULL);
+#else
 	SoupURI *failing_uri = soup_uri_new(failing_text);
+#endif
 
 	GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(browser->window->main_window),
 	                                           GTK_DIALOG_MODAL & GTK_DIALOG_DESTROY_WITH_PARENT,
@@ -397,6 +454,10 @@ WebViewCb_load_failed_with_tls_errors(WebKitWebView *web_view,
 	                                           failing_text);
 	gtk_dialog_add_buttons(
 	    GTK_DIALOG(dialog), _("Temporarily Add Exception"), 1, _("Continue"), 0, NULL);
+	if(!failing_uri)
+	{
+		gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog), 1, FALSE);
+	}
 	gtk_dialog_set_default_response(GTK_DIALOG(dialog), 0);
 	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), "%s\n", error_details);
 
@@ -404,12 +465,30 @@ WebViewCb_load_failed_with_tls_errors(WebKitWebView *web_view,
 
 	if(dialog_response == 1)
 	{
+#ifndef USE_LIBSOUP2
+		webkit_web_context_allow_tls_certificate_for_host(
+		    webkit_web_view_get_context(browser->webView), certificate, g_uri_get_host(failing_uri));
+#else
 		webkit_web_context_allow_tls_certificate_for_host(
 		    webkit_web_view_get_context(browser->webView), certificate, failing_uri->host);
+#endif
 		webkit_web_view_reload(browser->webView);
 	}
 
+#ifndef USE_LIBSOUP2
+	/* Calling g_free(failing_uri) ought to be the correct way but this causes a segfault.
+	 *
+	 * - documentation describes it as something which should be free
+	 * - implementation seems to make it a pointer that should be freed
+	 * - epiphany doesn't seems to free/unref it but gnome code seems to frequently have memleaks
+	 *
+	 * Decided to at least continue to try with using g_uri_unref(failing_uri) instead.
+	 * Related fediverse post: <https://queer.hacktivis.me/objects/cec7c4e8-6a58-4358-85bf-b66f9bb21a98>
+	 */
+	g_uri_unref(failing_uri);
+#else
 	soup_uri_free(failing_uri);
+#endif
 	g_free(error_details);
 	gtk_widget_destroy(dialog);
 
@@ -417,13 +496,14 @@ WebViewCb_load_failed_with_tls_errors(WebKitWebView *web_view,
 }
 
 static void
-web_contextCb_download_started(WebKitWebContext *web_context,
+web_contextCb_download_started(WebKitWebContext *UNUSED(web_context),
                                WebKitDownload *webkit_download,
                                gpointer user_data)
 {
-	(void)web_context;
 	struct Client *browser    = (struct Client *)user_data;
-	struct Download *download = malloc(sizeof(struct Client));
+	struct Download *download = malloc(sizeof(struct Download));
+
+	assert(webkit_download);
 
 	if(download != NULL)
 	{
@@ -490,38 +570,34 @@ auto_load_imagesCb_toggled(GtkButton *auto_load_images, gpointer user_data)
 }
 
 static void
-backCb_clicked(GtkButton *back, gpointer user_data)
+backCb_clicked(GtkButton *UNUSED(back), gpointer user_data)
 {
-	(void)back;
 	struct Client *browser = (struct Client *)user_data;
 
 	webkit_web_view_go_back(browser->webView);
 }
 
 static void
-forwardCb_clicked(GtkButton *forward, gpointer user_data)
+forwardCb_clicked(GtkButton *UNUSED(forward), gpointer user_data)
 {
-	(void)forward;
 	struct Client *browser = (struct Client *)user_data;
 
 	webkit_web_view_go_forward(browser->webView);
 }
 
 static void
-printCb_clicked(GtkButton *print, gpointer user_data)
+printCb_clicked(GtkButton *UNUSED(print), gpointer user_data)
 {
-	(void)print;
 	struct Client *browser = (struct Client *)user_data;
 
 	WebKitPrintOperation *print_operation = webkit_print_operation_new(browser->webView);
 
-	webkit_print_operation_run_dialog(print_operation, GTK_WINDOW(browser->window));
+	webkit_print_operation_run_dialog(print_operation, GTK_WINDOW(browser->window->main_window));
 }
 
 static gboolean
-SearchEntryCb_next__match(GtkSearchEntry *search, gpointer user_data)
+SearchEntryCb_next__match(GtkSearchEntry *UNUSED(search), gpointer user_data)
 {
-	(void)search;
 	struct Client *browser               = (struct Client *)user_data;
 	WebKitFindController *findController = webkit_web_view_get_find_controller(browser->webView);
 
@@ -531,9 +607,8 @@ SearchEntryCb_next__match(GtkSearchEntry *search, gpointer user_data)
 }
 
 static gboolean
-SearchEntryCb_previous__match(GtkSearchEntry *search, gpointer user_data)
+SearchEntryCb_previous__match(GtkSearchEntry *UNUSED(search), gpointer user_data)
 {
-	(void)search;
 	struct Client *browser               = (struct Client *)user_data;
 	WebKitFindController *findController = webkit_web_view_get_find_controller(browser->webView);
 
@@ -555,9 +630,8 @@ SearchEntryCb_search__changed(GtkSearchEntry *search, gpointer user_data)
 }
 
 static gboolean
-SearchEntryCb_stop__search(GtkSearchEntry *search, gpointer user_data)
+SearchEntryCb_stop__search(GtkSearchEntry *UNUSED(search), gpointer user_data)
 {
-	(void)search;
 	struct Client *browser               = (struct Client *)user_data;
 	WebKitFindController *findController = webkit_web_view_get_find_controller(browser->webView);
 
@@ -567,26 +641,26 @@ SearchEntryCb_stop__search(GtkSearchEntry *search, gpointer user_data)
 }
 
 static gboolean
-widgetCb_drop_button3_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
+widgetCb_drop_button3_event(GtkWidget *UNUSED(widget), GdkEvent *event, gpointer UNUSED(user_data))
 {
-	(void)widget;
-	(void)user_data;
-
 	// Button3 being right-click on right-handed mode, left-click on left-handed mode
 	return ((GdkEventButton *)event)->button == 3;
 }
 
 struct Client *
-new_browser(struct Window *window, const gchar *target_url, WebKitWebView *related_web_view)
+new_browser(struct Window *window, const gchar *target_url, struct Client *old_browser)
 {
 	struct Client *browser = malloc(sizeof(struct Client));
-	target_url             = badwolf_ensure_uri_scheme(target_url, (related_web_view == NULL));
+	target_url             = badwolf_ensure_uri_scheme(target_url, (old_browser == NULL));
 	char *badwolf_l10n     = NULL;
 
+	WebKitWebContext *web_context = NULL;
+
 	if(browser == NULL) return NULL;
 
-	browser->window = window;
-	browser->box    = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
+	browser->window     = window;
+	browser->context_id = old_browser == NULL ? context_id_counter++ : old_browser->context_id;
+	browser->box        = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
 	gtk_widget_set_name(browser->box, "browser__box");
 
 	browser->toolbar = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
@@ -625,24 +699,32 @@ new_browser(struct Window *window, const gchar *target_url, WebKitWebView *relat
 	browser->statuslabel = gtk_label_new(NULL);
 	gtk_widget_set_name(browser->statuslabel, "browser__statuslabel");
 
-	setenv("GTK_THEME", ":light", 0);
-
-	WebKitWebContext *web_context = webkit_web_context_new_ephemeral();
-	webkit_web_context_set_sandbox_enabled(web_context, TRUE);
-	webkit_web_context_set_process_model(web_context,
-	                                     WEBKIT_PROCESS_MODEL_MULTIPLE_SECONDARY_PROCESSES);
-
-	webkit_web_context_set_web_extensions_directory(web_context, web_extensions_directory);
-
-	badwolf_l10n = getenv("BADWOLF_L10N");
-
-	if(badwolf_l10n != NULL)
+	if(old_browser == NULL)
 	{
-		gchar **languages = g_strsplit(badwolf_l10n, ":", -1);
-		webkit_web_context_set_spell_checking_languages(web_context, (const gchar *const *)languages);
-		g_strfreev(languages);
+		WebKitWebsiteDataManager *website_data_manager = webkit_website_data_manager_new_ephemeral();
+		webkit_website_data_manager_set_itp_enabled(website_data_manager, TRUE);
 
-		webkit_web_context_set_spell_checking_enabled(web_context, TRUE);
+		web_context = webkit_web_context_new_with_website_data_manager(website_data_manager);
+		g_object_unref(website_data_manager);
+		webkit_web_context_set_sandbox_enabled(web_context, TRUE);
+		webkit_web_context_set_web_extensions_directory(web_context, web_extensions_directory);
+
+		g_signal_connect(G_OBJECT(web_context),
+		                 "download-started",
+		                 G_CALLBACK(web_contextCb_download_started),
+		                 browser);
+
+		/* flawfinder: ignore. Consider that g_strsplit is safe enough */
+		badwolf_l10n = getenv("BADWOLF_L10N");
+
+		if(badwolf_l10n != NULL)
+		{
+			gchar **languages = g_strsplit(badwolf_l10n, ":", -1);
+			webkit_web_context_set_spell_checking_languages(web_context, (const gchar *const *)languages);
+			g_strfreev(languages);
+
+			webkit_web_context_set_spell_checking_enabled(web_context, TRUE);
+		}
 	}
 
 	WebKitSettings *settings = webkit_settings_new_with_settings(BADWOLF_WEBKIT_SETTINGS);
@@ -656,12 +738,18 @@ new_browser(struct Window *window, const gchar *target_url, WebKitWebView *relat
 	                                                "web-context",
 	                                                web_context,
 	                                                "related-view",
-	                                                related_web_view,
+	                                                old_browser == NULL ? NULL : old_browser->webView,
 	                                                "settings",
 	                                                settings,
+	                                                "user-content-manager",
+	                                                window->content_manager,
 	                                                NULL));
+
 	gtk_widget_set_name(GTK_WIDGET(browser->webView), "browser__webView");
-	g_object_unref(web_context);
+	if(old_browser == NULL)
+	{
+		g_object_unref(web_context);
+	}
 	g_object_unref(settings);
 
 	gtk_box_pack_start(
@@ -720,6 +808,15 @@ new_browser(struct Window *window, const gchar *target_url, WebKitWebView *relat
 	gtk_label_set_single_line_mode(GTK_LABEL(browser->statuslabel), TRUE);
 	gtk_label_set_ellipsize(GTK_LABEL(browser->statuslabel), BADWOLF_STATUSLABEL_ELLIPSIZE);
 
+	if(bookmarks_completion_model != NULL)
+	{
+		GtkEntryCompletion *location_completion = gtk_entry_completion_new();
+		GtkTreeModel *location_completion_model = bookmarks_completion_model;
+
+		bookmarks_completion_setup(location_completion, location_completion_model);
+		gtk_entry_set_completion(GTK_ENTRY(browser->location), location_completion);
+	}
+
 	gtk_entry_set_text(GTK_ENTRY(browser->location), target_url);
 	gtk_entry_set_input_purpose(GTK_ENTRY(browser->location), GTK_INPUT_PURPOSE_URL);
 
@@ -787,26 +884,21 @@ new_browser(struct Window *window, const gchar *target_url, WebKitWebView *relat
 	                 "notify::estimated-load-progress",
 	                 G_CALLBACK(WebViewCb_notify__estimated_load_progress),
 	                 browser);
-	g_signal_connect(browser->webView, "create", G_CALLBACK(WebViewCb_create), window);
+	g_signal_connect(browser->webView, "create", G_CALLBACK(WebViewCb_create), browser);
 	g_signal_connect(browser->webView, "close", G_CALLBACK(WebViewCb_close), browser);
+
 	g_signal_connect(
 	    browser->webView, "key-press-event", G_CALLBACK(WebViewCb_key_press_event), browser);
 	g_signal_connect(browser->webView, "scroll-event", G_CALLBACK(WebViewCb_scroll_event), browser);
 	g_signal_connect(
 	    browser->webView, "permission-request", G_CALLBACK(WebViewCb_permission_request), NULL);
-	g_signal_connect(browser->webView, "decide-policy", G_CALLBACK(WebViewCb_decide_policy), NULL);
+	g_signal_connect(browser->webView, "decide-policy", G_CALLBACK(WebViewCb_decide_policy), browser);
 	g_signal_connect(browser->webView,
 	                 "load-failed-with-tls-errors",
 	                 G_CALLBACK(WebViewCb_load_failed_with_tls_errors),
 	                 browser);
 	g_signal_connect(browser->webView, "load-changed", G_CALLBACK(WebViewCb_load_changed), browser);
 
-	/* signals for WebView's WebContext */
-	g_signal_connect(G_OBJECT(web_context),
-	                 "download-started",
-	                 G_CALLBACK(web_contextCb_download_started),
-	                 browser);
-
 	/* signals for search widget */
 	g_signal_connect(browser->search, "next-match", G_CALLBACK(SearchEntryCb_next__match), browser);
 	g_signal_connect(
@@ -818,7 +910,7 @@ new_browser(struct Window *window, const gchar *target_url, WebKitWebView *relat
 	/* signals for box container */
 	g_signal_connect(browser->box, "key-press-event", G_CALLBACK(boxCb_key_press_event), browser);
 
-	if(related_web_view == NULL) webkit_web_view_load_uri(browser->webView, target_url);
+	if(old_browser == NULL) webkit_web_view_load_uri(browser->webView, target_url);
 
 	return browser;
 }
@@ -858,9 +950,8 @@ badwolf_new_tab(GtkNotebook *notebook, struct Client *browser, bool auto_switch)
 }
 
 static void
-new_tabCb_clicked(GtkButton *new_tab, gpointer user_data)
+new_tabCb_clicked(GtkButton *UNUSED(new_tab), gpointer user_data)
 {
-	(void)new_tab;
 	struct Window *window  = (struct Window *)user_data;
 	struct Client *browser = new_browser(window, NULL, NULL);
 
@@ -868,18 +959,19 @@ new_tabCb_clicked(GtkButton *new_tab, gpointer user_data)
 }
 
 static void
-closeCb_clicked(GtkButton *close, gpointer user_data)
+closeCb_clicked(GtkButton *UNUSED(close), gpointer user_data)
 {
-	(void)close;
 	struct Client *browser = (struct Client *)user_data;
 
 	webkit_web_view_try_close(browser->webView);
 }
 
 static void
-notebookCb_switch__page(GtkNotebook *notebook, GtkWidget *page, guint page_num, gpointer user_data)
+notebookCb_switch__page(GtkNotebook *notebook,
+                        GtkWidget *page,
+                        guint UNUSED(page_num),
+                        gpointer user_data)
 {
-	(void)page_num;
 	struct Window *window = (struct Window *)user_data;
 	GtkWidget *label      = gtk_notebook_get_tab_label(notebook, page);
 
@@ -887,15 +979,87 @@ notebookCb_switch__page(GtkNotebook *notebook, GtkWidget *page, guint page_num,
 	gtk_window_set_title(GTK_WINDOW(window->main_window), gtk_widget_get_tooltip_text(label));
 }
 
+void
+content_managerCb_ready(GObject *UNUSED(store), GAsyncResult *result, gpointer user_data)
+{
+	struct Window *window = (struct Window *)user_data;
+	GError *err           = NULL;
+
+	WebKitUserContentFilter *filter =
+	    webkit_user_content_filter_store_load_finish(window->content_store, result, &err);
+
+	if(filter == NULL)
+	{
+		if(err == NULL)
+		{
+			fprintf(stderr, _("badwolf: failed to load content-filter, err: [%d] %s\n"), -1, "unknown");
+		}
+		else
+		{
+			fprintf(stderr,
+			        _("badwolf: failed to load content-filter, err: [%d] %s\n"),
+			        err->code,
+			        err->message);
+		}
+	}
+	else
+	{
+		fprintf(stderr, _("badwolf: content-filter loaded, adding to content-manager…\n"));
+		webkit_user_content_manager_add_filter(window->content_manager, filter);
+	}
+}
+
+static void
+storeCb_finish(WebKitUserContentFilterStore *UNUSED(store),
+               GAsyncResult *result,
+               gpointer user_data)
+{
+	struct Window *window = (struct Window *)user_data;
+	GError *err           = NULL;
+
+	WebKitUserContentFilter *filter =
+	    webkit_user_content_filter_store_save_finish(window->content_store, result, &err);
+
+	if(filter == NULL)
+	{
+		if(err == NULL)
+		{
+			fprintf(stderr,
+			        _("badwolf: failed to compile content-filters.json, err: [%d] %s\n"),
+			        -1,
+			        "unknown");
+		}
+		else
+		{
+			fprintf(stderr,
+			        _("badwolf: failed to compile content-filters.json, err: [%d] %s\n"),
+			        err->code,
+			        err->message);
+		}
+	}
+	else
+	{
+		webkit_user_content_filter_store_load(
+		    window->content_store, "a", NULL, content_managerCb_ready, window);
+	}
+}
+
 int
 main(int argc, char *argv[])
 {
-	struct Window *window = &(struct Window){NULL, NULL, NULL, NULL};
+	struct Window *window = &(struct Window){NULL, NULL, NULL, NULL, NULL, NULL};
+	GApplication *application;
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, DATADIR "/locale");
 	bind_textdomain_codeset(PACKAGE, "UTF-8");
 	textdomain(PACKAGE);
 
+	application = g_application_new("me.hacktivis.badwolf",
+	                                G_APPLICATION_HANDLES_COMMAND_LINE |
+	                                    G_APPLICATION_SEND_ENVIRONMENT | G_APPLICATION_NON_UNIQUE);
+	g_application_register(application, NULL, NULL);
+	//g_application_activate(application);
+
 	gtk_init(&argc, &argv);
 
 	fprintf(stderr, _("Running Badwolf version: %s\n"), version);
@@ -914,10 +1078,31 @@ main(int argc, char *argv[])
 	    g_build_filename(g_get_user_data_dir(), "badwolf", "webkit-web-extension", NULL);
 	fprintf(stderr, _("webkit-web-extension directory set to: %s\n"), web_extensions_directory);
 
+	bookmarks_completion_model = bookmarks_completion_init();
+	g_object_ref(bookmarks_completion_model);
+
 	window->main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
 	window->notebook    = gtk_notebook_new();
 	window->new_tab = gtk_button_new_from_icon_name("tab-new-symbolic", GTK_ICON_SIZE_SMALL_TOOLBAR);
-	window->downloads_tab = badwolf_downloads_tab_new();
+	window->downloads_tab   = badwolf_downloads_tab_new();
+	window->content_manager = webkit_user_content_manager_new();
+
+	load_userscripts(window->content_manager);
+
+	gchar *contentFilterPath =
+	    g_build_filename(g_get_user_config_dir(), g_get_prgname(), "content-filters.json", NULL);
+	GFile *contentFilterFile = g_file_new_for_path(contentFilterPath);
+	fprintf(stderr, _("content-filters file set to: %s\n"), contentFilterPath);
+
+	gchar *filtersPath = g_build_filename(g_get_user_cache_dir(), g_get_prgname(), "filters", NULL);
+	window->content_store = webkit_user_content_filter_store_new(filtersPath);
+
+	webkit_user_content_filter_store_save_from_file(window->content_store,
+	                                                "a",
+	                                                contentFilterFile,
+	                                                NULL,
+	                                                (GAsyncReadyCallback)storeCb_finish,
+	                                                window);
 
 	gtk_window_set_default_size(
 	    GTK_WINDOW(window->main_window), BADWOLF_DEFAULT_WIDTH, BADWOLF_DEFAULT_HEIGHT);
@@ -925,6 +1110,7 @@ main(int argc, char *argv[])
 	gtk_window_set_icon_name(GTK_WINDOW(window->main_window), "badwolf");
 
 	gchar *provider_path_app = g_build_filename(DATADIR, "interface.css", NULL);
+	/* flawfinder: ignore, just a presence check */
 	if(access(provider_path_app, R_OK) == 0)
 	{
 		GtkCssProvider *css_provider_app = gtk_css_provider_new();
@@ -938,6 +1124,7 @@ main(int argc, char *argv[])
 
 	gchar *provider_path_user =
 	    g_build_filename(g_get_user_data_dir(), "badwolf", "interface.css", NULL);
+	/* flawfinder: ignore, just a presence check */
 	if(access(provider_path_user, R_OK) == 0)
 	{
 		GtkCssProvider *css_provider_user = gtk_css_provider_new();
@@ -981,6 +1168,8 @@ main(int argc, char *argv[])
 
 	gtk_main();
 
+	g_object_unref(bookmarks_completion_model);
+
 #if 0
 	/* TRANSLATOR Ignore this entry. Done for forcing Unicode in xgettext. */
 	_("ø");
